/*
  Originally, Frequ synth for TR-7200GII
  Denys VK3ZYZ (with help from Google and others)
  Fixes by Ray and Josh 20190728
  CTCSS tones added :)
  try to port it to OLED. (S,+,-)
  Add mode display
  Add CTCSS display *************** made it float. Check to see if tones still work! YES :)
  Invert frequ display on TX
  Had CTCSS on wrong channel Fixed!
  made display reset = -1

  Add CTCSS as per https://www.amateurradio.com.au/repeaters

  Max number of channels = 61 as RAM clashes.
  Mod for low band FM828
  Make it high side injection so coils closer to original
  Fix channel number.
  Adjust OLED freq display position.

  20211202 add calibration.
*/

#define Version "FM828E_VFO_OLED20211202"

#include "si5351.h"
#include "Wire.h"
#include "Adafruit_GFX.h"
#include "Adafruit_SSD1306.h"
#define SCREEN_WIDTH 128 // OLED display width, in pixels
#define SCREEN_HEIGHT 32 // OLED display height, in pixels

// Declaration for an SSD1306 display connected to I2C (SDA, SCL pins)
#define OLED_RESET     -1 // Reset pin # (or -1 if sharing Arduino reset pin)
Adafruit_SSD1306 display(SCREEN_WIDTH, SCREEN_HEIGHT, &Wire, OLED_RESET);

// synthesiser default settings on start up.
unsigned long RXgenfrequency  = 53725000;
unsigned long TXgenfrequency  = 52725000;
unsigned long IFoffset        = 10700000;    // IF offset
unsigned long TXmultiplier    = 8;           // transmitter frequency multiplier
unsigned long RXmultiplier    = 2;           // receiver frequency multiplier.

float         txDisplayFreq   = 53.725;
float         rxDisplayFreq   = 52.725;
int           mode            = HIGH;        // HIGH = normal, LOW = REV
int           TRXmode         = LOW;        // HIGH = TX, LOW = RX
int           channel         = 30;           // start on channel number "31" (starts from 0)
int encoder;                                 // encoder direction: CCW=-1; NC=0; CW=1
int timer = 100;           // sub tone setting. The higher the number, the slower the timing.
int  i ;

int32_t cal_factor = 30151; //  <<<--- found from Si5351 calibrator sketch.

Si5351 clockgen;                             // initalize clockgen

// encoder definitions
#define ENCODER_A     A0
#define ENCODER_B     A1
//#define SSB_SWITCH  A2           // HIGH = LSB, LOW = USB.
//#define AM_SWITCH   A3           // not used yet
// A4 = SDA
// A5 = SCL
#define AnalogIn1     A6           // not used yet
#define AnalogIn2     A7           // not used yet
#define spare         A2           // not used yet
#define TRX            2           // input HIGH = TX, LOW = RX.
#define Input2         3           // not used yet
#define Output1        4           // not used yet
#define REV_SWITCH     5           // Encoder Button  repeater reverse
#define spareSW        6           // not used yet

/* ********************* PortB0-PB5 used to generate CTCSS tones via R2R nerwork.
#define spare1         7           // Sub Tone R2R b0 PB0
#define spare2         8           // Sub Tone R2R b1 PB1
#define spare3         9           // Sub Tone R2R b2 PB2
#define spare4        10           // Sub Tone R2R b3 PB3
#define spare5        11           // Sub Tone R2R b4 PB4
#define spare6        12           // Sub Tone R2R b5 PB5
*/

#define Output2       13           // not used yet

#define ENCODER_INPUT_PORT PINC
#define On  1
#define Off 0
#define RefFrequ      500000        // 16Mhz / 32 steps for CTCSS sine wave generator
int     SineDivider = 4065;         // default sine wave divider

  
long int RXFrequs []      // Edit this for RX frequencies wanted.
{ 52525,  // chan 1
  53000,
  53025,
  53050,
  53075,
  53100,
  53125,
  53150,
  53175,
  53200,  // chan 10
  53225,
  53250,
  53275,
  53300,
  53325,
  53350,
  53375,
  53400,
  53425,
  53450,  // chan 20
  53475,
  53500,
  53525,

  53550, 
  53575,
  53600,
  53625,
  53650,
  53675,
  53700,  // chan 30
  53725,
  53750,
  53775,
  53800,
  53825,
  53850,
  53875,
  53900,
  53925,
  53950,  // chan 40
  53975, 

};

int TXFrequs []      // Edit this for TX frequencies wanted. Added to RX frequs.
{ 0,  // chan 1
  0,
  0,
  0,
  0,
  0,
  0,
  0,
  0,
  0,  // chan 10
  0,
  0,
  0,
  0,
  0,
  0,
  0,
  0,
  0,
  0,  // chan 20
  0,
  0,
  0,

  -1000, 
  -1000, 
  -1000,
  -1000,
  -1000,
  -1000,
  -1000,  // chan 30
  -1000,
  -1000,
  -1000,
  -1000,
  -1000,
  -1000,
  -1000,
  -1000,
  -1000,
  -1000,  // chan 40
  -1000,
 
};

float SubTone []      // Edit this for CTCSS frequencies wanted.
{ 0,      // "0 ="  chan 1
  0,
  0,
  0,
  0,
  0,
  0,
  0,
  0,
  0,  // chan 10
  0,
  0,
  0,
  0,
  0,
  0,
  0,
  0,
  0,
  0,  // chan 20
  0,
  0,
  0,

  0, 
  91.5,
  0,
  88.5,
  91.5,   
  0,
  0,  // chan 30
  123.0,
  0,
  0,
  0,
  0,
  91.5,   
  0,
  0,
  0,
  0,  // chan 40
  0,   
 
};

int sinetable[] // one cycle sine wave via R2R network.
{ 0,1,2,3,5,7,10,13,16,19,22,25,27,28,29,30,31,30,29,28,27,25,22,19,16,13,10,7,5,3,2,1, };


void setup() {
  // SSD1306_SWITCHCAPVCC = generate display voltage from 3.3V internally
  if(!display.begin(SSD1306_SWITCHCAPVCC, 0x3C)) { // Address may be 0x3D 
//    Serial.println(F("SSD1306 allocation failed"));
    for(;;); // Don't proceed, loop forever
  }
  display.clearDisplay();
  display.display();                  // Update screen
  display.setTextSize(1);             // Normal 1:1 pixel scale
  display.setTextColor(WHITE,BLACK);  // Draw white text
  display.setCursor(0,0);             // Start at top-left corner
  display.println(F(Version));
  display.display(); // Update screen
  delay(1000);
  display.clearDisplay();
  display.display();                  // Update screen
  delay(100);
  display.setCursor(0,0);
  display.print("Ch                    ");
  display.display(); // Update screen
  bool i2c_found;

  // Start serial and initialize the Si5351
  i2c_found = clockgen.init(SI5351_CRYSTAL_LOAD_8PF, 0, 0);
  clockgen.set_correction(cal_factor, SI5351_PLL_INPUT_XO); // <<<<<<<this is found from Si5351 calibrator code.


  // configure rotary encoder & push-button
  pinMode(ENCODER_A, INPUT_PULLUP);     // encoder leads
  pinMode(ENCODER_B, INPUT_PULLUP);
  pinMode(REV_SWITCH, INPUT_PULLUP);    // Hi = nomal, LOW =  Reverse
  pinMode(TRX, INPUT_PULLUP);     // Hi = TX, LOW =  RX
  pinMode(Input2, INPUT_PULLUP);
  pinMode(Output1, OUTPUT);
  DDRB  = B111111;                // sets Arduino pins 7 to 12 as outputs.
  PORTB = B100000;                // R2R out at rest to half level.
  pinMode(Output2, OUTPUT);
  digitalWrite(Output1, HIGH);    // output1 off
  digitalWrite(Output2, HIGH);    // output2 off

  encoder = read_encoder();       // returns +1/0/-1 (cw/nc/ccw)
  mode=( digitalRead(REV_SWITCH));
  TRXmode=(digitalRead(TRX));
    set_synth(mode, TRXmode);
}

/******************************
  main loop
*******************************/
void loop() {
  if ( digitalRead(REV_SWITCH) != mode){
    mode=( digitalRead(REV_SWITCH));
    set_synth(mode, TRXmode);
  }
  if (digitalRead(TRX) !=TRXmode){
    TRXmode=(digitalRead(TRX));
    set_synth(mode, TRXmode);  
  }
  // check the frequency
  encoder = read_encoder();                       // returns +1/0/-1 (cw/nc/ccw)
  if (encoder != 0) {
    channel = channel + encoder;
    channel = constrain(channel, 0, 40);          // max 41 channels.
    set_synth(mode, TRXmode);
  }
}

/******************************
  read encoder direction
*******************************/
int read_encoder() {
  // definitions
#define DELAY 100                               // number of polling-loops   
  static unsigned short integrator;               // encoder debounce

  /* the array values are determined by combining the previous and current A1,A0
    Gray Code bit-patterns to form an array-index. +1 is assigned to all clockwise
    (CW) indexes; -1 is assigned to all counter-clockwise (CCW) indexes; and 0 is
    assigned to the remaining positions to indicate no movement*/
  static char encoder[] = {0, -1, 1, 0, 1, 0, 0, -1, -1, 0, 0, 1, 0, 1, -1, 0}; // SWAPPED PHASES
  static unsigned char encoder_index = 0;
  static long count = 0;                         // CW rotation = +1; CCW rotation = -1
  static long this_reading;
  static long last_reading;
  char value;                                    // return value

  /* check for encoder rotation. The absolute "count" value changes as the encoder rotates.
    Typically there are four states (counts) per click, but often more due to contact bounce.
    Bounce doesn't matter so long as the "count" decreases" for CCW rotation; remains constant
    when there is no rotation; and increases with CW rotation. Bounce can be completely
    eliminated if we look for the "detent" code and integrate as shown below. */
  encoder_index <<= 2;                           // shift previous bit-pattern left two places
  encoder_index |= (ENCODER_INPUT_PORT & 0x03);  // get current bit-pattern and combine with previous
  count += encoder[( encoder_index & 0x0f )];

  // debounce encoder
  if ((ENCODER_INPUT_PORT & 0x03) == 3) {        // encoder on an indent (3 decimal == 00000011 binary)
    if (integrator < DELAY) {
      integrator++;
    }
  } else {                                       // encoder not resting on indent
    if (integrator > 0) {
      integrator--;
    }
  }

  // prepare the "return" value
  if (integrator >= DELAY) {                     // encoder deemed to be stationary
    this_reading = count;
    if (this_reading < last_reading) value = -1;
    if (this_reading == last_reading) value = 0;
    if (this_reading > last_reading) value = 1;
    last_reading = this_reading;                 // both readings now hold current "count"
    integrator = 0;                              // reset integrator
  } else {
    value = 0;
  }
  return (value);
}

/*******************************
  set the Si5351 CLK0 frequency
 ******************************/
 
void set_synth( int frmode, int trx )
{  
  display.setTextSize(1);     // Normal 1:1 pixel scale
  display.setCursor(18,0);   // write channel number to the LCD 60
  display.print("  ");        // clear the old number
  display.setCursor(18,0);   // write channel number to the LCD
  if (channel < 9)            // "9" = 10 as array starts at 0
  {
    display.print(" ");       // extra space for formalling a single digit.
  }
  display.print((channel + 1)); // display the channel number

  display.setCursor(34,0);   // write mode to the LCD
  if (TXFrequs[channel]== 0)
  {display.print(" SIM "); }
    if (TXFrequs[channel] == 1000)
  {display.print("+1000"); }
    if (TXFrequs[channel] ==-1000)
  {display.print("-1000"); }

  display.setCursor(70,0);          // write CTCSS to the LCD
  display.print("         ");
  if((SubTone[channel])!=0) 
  {display.setCursor(70,0); 
   display.print("tone");
    if((SubTone[channel])<100)        // pad leading spacee if needed
  {  display.print(" ");}             
  display.print((SubTone[channel]),1);  
  }
// Select FWD or REV frequencies
  if (frmode == HIGH)    // test for fwd or rev repeater...
  { RXgenfrequency = (RXFrequs[channel]); // normal operation
    TXgenfrequency = ((RXFrequs[channel]) + (TXFrequs[channel]));
  }
  else
  {
    TXgenfrequency = (RXFrequs[channel]); // reverse repeater operation
    RXgenfrequency = ((RXFrequs[channel]) + (TXFrequs[channel]));
  }

// Then display them
  if ( trx == LOW) // =RX.
  {  
  rxDisplayFreq = float (RXgenfrequency)/1000;  // display the RX frequency
  display.setCursor(10,10);                      // start of second LCD line
  display.setTextSize(3);                       // Normal 1:1 pixel scale
  display.print(rxDisplayFreq, 3);
  }
   if ( trx == HIGH) // =TX.
  {
  display.setTextColor(BLACK,WHITE);            // Invert Frequ Display
  display.setCursor(10,10);
  display.setTextSize(3);                       // large text
  txDisplayFreq = float (TXgenfrequency)/1000;
  display.print(txDisplayFreq, 3);
  }
  display.display(); // Update screen
  display.setTextColor(WHITE,BLACK);            // back to normal text

    display.display(); // Update screen
// calculate the actual frequencies to set

  RXgenfrequency = (((RXgenfrequency * 1000) + IFoffset) / RXmultiplier);
  TXgenfrequency = ((TXgenfrequency  * 1000) / TXmultiplier);

// is it transmittig or receiving?

  if ( trx == LOW) // =RX.
  { 
    clockgen.output_enable(SI5351_CLK0, 1);                 // turn RX gen on
    clockgen.set_freq((RXgenfrequency * 100), SI5351_CLK0); // set the RX generator frequ,
    clockgen.output_enable(SI5351_CLK1, 0);                 // turn TX gen off
    TIMSK1 = 0;                                             // subtone off
  }
  else                            // =TX.
  { 
    clockgen.output_enable(SI5351_CLK1, 1);                 // turn TX gen on
    clockgen.set_freq((TXgenfrequency * 100), SI5351_CLK1); // set the TX generator frequ,
    clockgen.output_enable(SI5351_CLK0, 0);                 // turn RX gen off
    if ((SubTone[channel]) != 0)
    { SineDivider = (RefFrequ/(SubTone[channel])); 
    CTCSS_Tone();                                           // subtone on
    }
    else 
    {
    TIMSK1 = 0;                                             // subtone off
    }
  }
}

void CTCSS_Tone()
{
  cli();                    // interupsts off for setup
  TCCR1A = 0;               // 
  TCCR1B = 0; //B00000010;       //  no prescaler
   // set compare match register for 1hz increments
  OCR1A = SineDivider ;// = (16*10^6) / (1*1024) - 1 (must be <65536)
  // turn on CTC mode
  TCCR1B |= (1 << WGM12);
  // Set CS10 no prescaler
  TCCR1B |= (0 << CS12) | (0 << CS11) | (1 << CS10);  
  // enable timer compare interrupt
  TIMSK1 |= (1 << OCIE1A);
  sei();                    // interupsts on      
}

ISR (TIMER1_COMPA_vect) {
  PORTB=(sinetable[i++]);   // write value from table to PORTB
  if(i>=32){
    i=0;
  }
}
